
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <unistd.h>
#include <pty.h>
#include <utmp.h>
#include <sys/wait.h>
#include <pthread.h>
#include <gif_lib.h>

#include "eink.h"


int FONT_MONO=1;

FT_Library     library;

FT_UInt glyph_index;
FT_Glyph glyph;
FT_BitmapGlyph glyph_bitmap;
FT_Glyph copy_glyph;

FTC_ScalerRec scaler;
FT_Face face;


FTC_Manager    cache_man;
FTC_CMapCache  cmap_cache;
FTC_ImageCache image_cache;
FTC_FaceID     face_id;
FTC_ImageTypeRec font_type;
FT_Int         cmap_index;

int FONT_WIDTH;
int FONT_HEIGHT;
int FONT_LINEHEIGHT;
int HOFFSET;

char *frame_buf;
char *frame_buf_bg;
char *frame_buf_bg_last;

int update_flag;
int update_flag_last;

int fb_update;

int eink_ignore = 0;;

pthread_t updater;

FT_Error face_requester( FTC_FaceID face_id, FT_Library library, FT_Pointer request_data, FT_Face* aface )
{
	FT_Error ret;
	FT_Face face;

	if( ( ret = FT_New_Face( library, (char*)face_id, 0, &face ) ) != FT_Err_Ok ){
		return ret;
	}
	if( ( ret = FT_Select_Charmap( face, FT_ENCODING_UNICODE ) ) != FT_Err_Ok ){
		return ret;
	}

	*aface = (FT_Face)face;
	return ret;
}

void eink_clear_lines( char *graphic_buf, int y, int height, attr_t attr )
{
	char *pos;
	unsigned char fg, bg, ul;
	if( height < 0 ) return;
	
	if( y == -1 ){
		memcpy( frame_buf, graphic_buf, HOFFSET * 600 / 2 );
		return;
	}

	attr_extract( attr, &fg, &bg, &ul );
	
	pos = frame_buf + y * FONT_LINEHEIGHT * 600 / 2 + ( HOFFSET * 600 / 2 );
	if( bg == 0 ){
		memcpy( pos, graphic_buf + y * FONT_LINEHEIGHT * 600 / 2 + ( HOFFSET * 600 / 2 ), 
					height * FONT_LINEHEIGHT * 600 / 2 );
	}else{
		int i;
		unsigned char buf;

		buf = bg | bg << 4;
		for( i = 0; i < height * FONT_LINEHEIGHT * 600 / 2; i ++ ){
			*pos = buf;
			pos ++;
		}
	}
}

void eink_clear_chars( char *graphic_buf, int x, int y, int width, attr_t attr )
{
	int i;
	char *pos;
	unsigned char fg, bg, ul;
	if( width < 0 ) return;

	attr_extract( attr, &fg, &bg, &ul );

	for( i = 0; i < FONT_LINEHEIGHT; i ++ ){
		pos = frame_buf + ( y * FONT_LINEHEIGHT + i ) * 600 / 2 + x * FONT_WIDTH / 2 + ( HOFFSET * 600 / 2 );
		if( bg == 0 ){
			memcpy( pos, graphic_buf + ( y * FONT_LINEHEIGHT + i ) * 600 / 2 + x * FONT_WIDTH / 2 + ( HOFFSET * 600 / 2 ), 
						width * FONT_WIDTH / 2 );
			pos += width * FONT_WIDTH / 2;
		}else{
			int i;
			unsigned char buf;

			buf = bg | bg << 4;
			for( i = 0; i < width * FONT_WIDTH / 2; i ++ ){
				*pos = buf;
				pos ++;
			}
		}
	}
}

void eink_gifprint( char *graphic_buf, GifFileType *gif, int x0, int y0, int num )
{
	int y;
	GifRecordType rectype;
	GifByteType *ext;
	static ColorMapObject *ColorMap;
	int code;
	int i;
    
	do{
		if( DGifGetRecordType( gif, &rectype ) == GIF_ERROR ){
			printf("Error loading gif image on DGifGetRecordType.\n");
		}
		if( rectype == EXTENSION_RECORD_TYPE ){
            if(DGifGetExtension( gif, &code, &ext ) == GIF_ERROR ){
				printf("Error loading gif image on DGifGetExtension.\n");
            }
			do{
				if( DGifGetExtensionNext( gif, &ext ) == GIF_ERROR ){
					printf("Error loading gif image on DGifGetExtensionNext.\n");
					break;
				}
			}while( ext );
		}else if( rectype == IMAGE_DESC_RECORD_TYPE ){
			GifPixelType *buf;
			ColorMap = ( gif->Image.ColorMap ? gif->Image.ColorMap : gif->SColorMap );
			if( ColorMap == NULL ){
				printf("Error loading gif image colormap.\n");
			}
			//printf("%dx%d ", gif->SWidth,gif->SHeight);
			buf = malloc( gif->SWidth * sizeof(GifPixelType) );
			if( DGifGetImageDesc( gif ) == GIF_ERROR ){
				printf("Error loading gif image on DGifGetImageDesc.\n");
			}
			for( y = 0; y < gif->SHeight; y ++ ){
				int x;
				if( DGifGetLine( gif, buf, gif->SWidth ) == GIF_ERROR ){
					printf("Error loading gif image on DGifGetLine.\n");
				}
				for( i = 0; i < num; i ++ ){
					for( x = 0; x < gif->SWidth; x ++ ){
						int pos;
						int color;
						
						color = 15 - ColorMap->Colors[ buf[ x ] ].Red / 16;
						pos = ( y0 + y ) * 600 + ( x0 + x + i * gif->SWidth );
						if( pos % 2 == 0 ){
							graphic_buf[ pos / 2 ] &= 0xF0;
							graphic_buf[ pos / 2 ] |= color;
						}else{
							graphic_buf[ pos / 2 ] &= 0x0F;
							graphic_buf[ pos / 2 ] |= color << 4;
						}
					}
				}
			}
			free( buf );
		}
	}while( rectype != TERMINATE_RECORD_TYPE );
}
void *updater_therad( void *data )
{
	int fb_update;
	int fb_hw;
	int i;
	char up[64];
	
	fb_hw = open( "/dev/fb0", O_WRONLY );
	if( fb_hw == -1 ) return 0;
	
	fb_update = open( "/proc/eink_fb/update_display", O_WRONLY | O_NONBLOCK );
	if( fb_update ==-1 ) return 0;
	
	printf( " Updater thread initialized.\n" );
	while( 1 ){
		if( update_flag_last ){
			update_flag = update_flag_last;
			update_flag_last= 0;
			if( !eink_ignore ){
				int start, end;
				start = 300 * 800; end = 0;

				if( update_flag == 7 ){
					int x1, x2, y1, y2;
					x1 = 300; x2 = 0;
					y1 = 800; y2 = 0;
					{
						char *bg;
						char *last;
						
						bg = frame_buf_bg;
						last = frame_buf_bg_last;
						for( i = 0; i < 800*600/2; i ++ ){
							if( *bg != *last ){
								*bg = *last;
								if( i < start ) start = i;
								if( i > end ) end = i;
								if( i / 300 < y1 ) y1 = i / 300;
								if( i / 300 > y2 ) y2 = i / 300;
								if( i % 300 < x1 ) x1 = i % 300;
								if( i % 300 > x2 ) x2 = i % 300;
							}
							bg ++; last ++;
						}
						end ++;
					}
					sprintf( up, "%d %d 0 1 %d %d %d %d", update_flag, 0/*fx:update_partical*/, x1, y1, x2, y2 );
				//	printf("(%d,%d)-(%d,%d)\n",x1, y1, x2, y2);
				}else{
					end = 300 * 800;
					start = 0;
					memcpy( frame_buf_bg, frame_buf_bg_last, 800 * 600 / 2 );
					sprintf( up, "%d", update_flag );
				}
				
				lseek( fb_hw, start, SEEK_SET );
				write( fb_hw, frame_buf_bg + start, end - start );
			}
			update_flag = 0;
			
			if( !eink_ignore ){
			//	printf("\nupdate with %s\n",up);
				write( fb_update, up, strlen(up) );
			}
		}else{
			usleep( 50000 );
		}
	}
	return NULL;
}

void eink_flush( int type, int option )
{
	if( type != 1 && type != 2 && type != 7 ){
		char up[16];
		update_flag = 0;
		if( !eink_ignore ){
			sprintf( up, "%d %d", type, option );
			write( fb_update, &up, strlen( up ) );
			fsync( fb_update );
		}
	}else{
		memcpy( frame_buf_bg_last, frame_buf, 800 * 600 / 2 );
		update_flag_last = type;
	}
}

void eink_ignore_flush( int ignore )
{
	eink_ignore = ignore;
}

void eink_cursor( char *graphic_buf, int x, int y, char bar, attr_t attr )
{
	int i, j;
	char buf;
	char *pos;
	unsigned char fg, bg, ul;
	

	attr_extract( attr, &fg, &bg, &ul );
	
	for( j = 0; j < 1; j ++ ){
		pos = frame_buf + ( ( y + 1 ) * FONT_LINEHEIGHT - 1 + j ) * 600 / 2 + ( x * FONT_WIDTH ) / 2 + ( HOFFSET * 600 / 2 );
		for( i = 0; i < FONT_WIDTH / 2; i ++ ){
			if( bar ){
				buf = fg | fg << 4;
			}else{
				if( bg == 0 ){
					buf = graphic_buf[ ( ( y + 1 ) * FONT_LINEHEIGHT - 1 ) * 600 / 2 + ( x * FONT_WIDTH ) / 2 + i + ( HOFFSET * 600 / 2 ) ];
				}else{
					buf = bg | bg << 4;
				}
			}
			*pos = buf;
			pos ++;
		}
	}
}

void eink_putc( char *graphic_buf, int x, int y, unsigned long key, attr_t attr )
{
	eink_putc2( graphic_buf, x * FONT_WIDTH, y * FONT_LINEHEIGHT + HOFFSET, FONT_WIDTH * 2, FONT_HEIGHT, FONT_LINEHEIGHT, key, attr );
}

void eink_putc2( char *graphic_buf, int x, int y, int w, int h, int lh, unsigned long key, attr_t attr )
{
	int i, j;
	char *pos;
	unsigned long code;
	unsigned char fg, bg, ul;
	unsigned char buf;
	int i2;
	int width;
	

	attr_extract( attr, &fg, &bg, &ul );

	code = key;
	if( key == '\\' ){
		code = '/';
	}
	if( FTC_Manager_LookupFace( cache_man, face_id, &face ) ){
		printf("Font cache error 1\n");
	}
	scaler.face_id = face_id;
	scaler.width = w;
	scaler.height = h;
	scaler.pixel = 1;
	
	FT_Size font_size;
	if( FTC_Manager_LookupSize( cache_man, &scaler, &font_size ) ){
		printf("Font cache error 2\n");
	}

	font_type.face_id = face_id;
	font_type.width   = 0;
	font_type.height  = h;
	font_type.flags   = FT_LOAD_RENDER | FT_LOAD_TARGET_NORMAL;
	
	cmap_index = FT_Get_Charmap_Index( face->charmap );
/*
	glyph_index = FT_Get_Char_Index( face, key ); 
	FT_Load_Glyph( face, glyph_index, FT_LOAD_DEFAULT | FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH );
	FT_Render_Glyph( face->glyph, FT_RENDER_MODE_NORMAL );
*/
	glyph_index = FTC_CMapCache_Lookup( cmap_cache, face_id, cmap_index, code );
	if( FTC_ImageCache_Lookup( image_cache, &font_type, glyph_index, &glyph, NULL ) ){
		printf("Font cache error 3\n");
	}
	if( glyph->format == FT_GLYPH_FORMAT_BITMAP ){
		glyph_bitmap = (FT_BitmapGlyph)glyph;
	}else{
		FT_Glyph_Copy( glyph, &copy_glyph );
		FT_Glyph_To_Bitmap( &copy_glyph, FT_RENDER_MODE_NORMAL, 0, 1);//FT_RENDER_MODE_NORMAL
		glyph_bitmap = (FT_BitmapGlyph)copy_glyph;
	}
//	FT_Render_Glyph( face->glyph, FT_RENDER_MODE_NORMAL );
	if( attr & ATTR_BOLD ){
		FT_Glyph_Copy( glyph, &copy_glyph );
		glyph_bitmap = (FT_BitmapGlyph)copy_glyph;
		FT_Bitmap_Embolden( library, &glyph_bitmap->bitmap, 32, 32 );
	}
	int top, left;
	
	top = glyph_bitmap->top;
//	printf("%c(%d,%d) ",key,top,glyph_bitmap->bitmap.rows);
	if( top < 0 ) top = 0;
	left = glyph_bitmap->left;
	for( j = 0; j < lh; j ++ ){
		pos = frame_buf + ( j + y ) * 600 / 2 + x / 2;
		buf = 0;
		i2 = 0;

		width = w / 2;
		if( key >= 0x80 ) width *= 2;
		for( i = 0; i < width; i ++ ){
			unsigned short color;
			buf = buf << 4;
			if( ( x + i ) / 2 >= 600 / 2 ) continue;
			if( left <= i && i < left + glyph_bitmap->bitmap.width &&
					( lh - top ) - ( lh - h ) <= j && 
					j < ( lh - top ) - ( lh - h ) + glyph_bitmap->bitmap.rows  ){
				if( key == '\\' ){
					color = glyph_bitmap->bitmap.buffer[ 
						( j - ( lh - top ) + ( lh - h ) ) * glyph_bitmap->bitmap.pitch + 
						( glyph_bitmap->bitmap.width - i - 1 ) - left ];
				}else{
					color = glyph_bitmap->bitmap.buffer[ 
						( j - ( lh - top ) + ( lh - h ) ) * glyph_bitmap->bitmap.pitch + 
						i - left ];
				}
				if( fg > bg ){
					if( !FONT_MONO ){
						color = color * fg / 255;
					}else{
						if( color > 160 ){
							color = 15;
						}else{
							color = 0;
						}
					}
				}else{
					if( !FONT_MONO ){
						color = fg + ( 255 - color ) * ( 15 - fg ) / ( 15 * 17 );
					}else{
						if( color > 160 ){
							color = 0;
						}else{
							color = 15;
						}
					}
				}
			}else{
				if( fg > bg ){
					color = 0;
				}else{
					color = 15;
				}
			}
			buf |= color;
			if( i2 % 2 == 1 ){
				unsigned char buf2;

				if( j == lh - 2 && ul != 0 ){
						buf2 = ul | ul << 4;
				}else if( bg == 0 ){
					buf2 = graphic_buf[ ( j + y ) * 600 / 2 + ( x + i ) / 2 ];
				}else{
					buf2 = bg | bg << 4;
				}
				if( fg > bg ){
					buf2 = maximum( buf2 & 0xF0, buf & 0xF0 ) | maximum( buf2 & 0xF, buf & 0xF );
				}else{
					buf2 = minimum( buf2 & 0xF0, buf & 0xF0 ) | minimum( buf2 & 0xF, buf & 0xF );
					// fg 5-0 bg 15
				}
				*pos = buf2;
				pos ++;
			}
			i2 ++;
		}
	}
//	printf("\n");
}

int eink_reinit( char *graphic_buf, int width, int height, int lineheight, int hoffset )
{
	FONT_LINEHEIGHT = lineheight;
	FONT_WIDTH = width;
	FONT_HEIGHT = height;
	HOFFSET = hoffset;
	
	memcpy( frame_buf, graphic_buf, 800 * 600 / 2 );

	return 1;
}

int eink_init( char *graphic_buf, char *fontname, int width, int height, int lineheight, int hoffset )
{
	face_id = fontname;
	FT_Init_FreeType( &library );
/*
	if( FT_New_Face( library, "/mnt/base-us/hackindle/DFJHSGW5.TTC", 0, &face ) ){
		return 0;
	}
	FT_Set_Pixel_Sizes( face, FONT_WIDTH * 2, FONT_HEIGHT );
*/
	
	FTC_Manager_New( library, 0, 0, 0, face_requester, NULL, &cache_man );
	FTC_CMapCache_New( cache_man, &cmap_cache );
	FTC_ImageCache_New( cache_man, &image_cache );
	
	frame_buf = malloc( 800 * 600 / 2 );
	frame_buf_bg = malloc( 800 * 600 / 2 );
	frame_buf_bg_last = malloc( 800 * 600 / 2 );
	
	fb_update = open( "/proc/eink_fb/update_display", O_WRONLY );
	if( fb_update ==-1 ){
		fprintf( stderr, "Failed to open update_display.\n" );
		return 0;
	}

	update_flag = 0;
	update_flag_last = 0;
	pthread_create( &updater, NULL, updater_therad, NULL );

	eink_reinit( graphic_buf, width, height, lineheight, hoffset );

	return 1;
}

void eink_font_mono( int mono )
{
	FONT_MONO = mono;
}
