#define _XOPEN_SOURCE
#define _BSD_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <unistd.h>
#include <pty.h>
#include <utmp.h>
#include <sys/wait.h>
#include <gif_lib.h>

#include "eink.h"
#include "hackindle.h"
#include "kindle-keypad.h"
#include "terminal.h"


#define LOGIN_CMD		("/bin/login")
#define MAX_BUFFER		16

#define FONT_HEIGHT_DEFAULT		16
#define FONT_WIDTH		(FONT_HEIGHT/2)
#define FONT_LINEHEIGHT	(FONT_HEIGHT+FONT_HEIGHT/4)
#define FONT_PATH		("/mnt/base-us/hackindle/MMCedar.ttf")

#define BACKGROUND_PATH	("/mnt/base-us/hackindle/kindle.img")
#define TOPBAR_LEFT_CAP		("/opt/amazon/ebook/img/ui/TopBar_LeftCap.gif")
#define TOPBAR_RIGHT_CAP	("/opt/amazon/ebook/img/ui/TopBar_RightCap.gif")
#define TOPBAR_MIDTEXTURE	("/opt/amazon/ebook/img/ui/TopBar_MidTexture.gif")
#define APPLICATION_NAME	("Hackindle Terminal")

struct winsize winsz;
struct terminal term;
struct position subcursor;
struct position startpos;
int stop = 0;
pid_t pid;


void ctrlc( int num )
{
	if( stop == 2 ){
		kill( pid, SIGKILL );
		exit(0);
	}else{
		stop = 1;
	}
}

unsigned long *get_selected( int *len )
{
	int s, e;
	
	s = startpos.y * winsz.ws_col + startpos.x;
	e = subcursor.y * winsz.ws_col + subcursor.x;
	if( s > e ){
		int tmp;
		tmp = e;
		e = s;
		s = tmp;
	}
	*len = e - s + 1;
	
	return &term.buffer[ term.ibuffer_disp ].text[ s ];
}

int main( int argc, char *argv[] )
{
	int hack   = 0;
	char *background_graphic;
	int FONT_HEIGHT;
	int i;
	int pty, pts;
	int cur_move = 0;
	fd_set key_rfds;
	int fd_max;
	int kev;
	struct timeval timeout;
	pid_t childs[5];
	int termnum = 0;
	int bar_height;

	for( i = 0; i < 5; i ++ ){
		childs[i] = fork();
		if( childs[i] == 0 ) break;
		usleep( 50000 );
	}
	termnum = i;
	if( termnum == 5 ){
		terminal_init( &term, 1, 1, 1 );
		signal( SIGINT, ctrlc );
		kev = kkeypad_init( fileno(stdout), get_selected, &term );
		kkeypad_ignore( 1 );
		hack = 0;
		while( !stop ){
			enum EVENT event;
			
			FD_ZERO( &key_rfds );
			FD_SET( kev, &key_rfds );
			timeout.tv_sec = 1;
			timeout.tv_usec = 0;
			fd_max = kev + 1;
			if( select( fd_max, &key_rfds, NULL, NULL, &timeout ) > 0 ){
				if( FD_ISSET( kev, &key_rfds ) ){
					read( kev, &event, sizeof( event ) );

					if( EVENT_SWITCHING1 <= event && event <= EVENT_SWITCHING5 && hack == 0 ){
						printf( "Resuming hackindle\n" );
						system( "dbus-send --system /default com.lab126.powerd.goingToScreenSaver" );
						usleep( 800000 );
						system( "killall -STOP cvm" );
						hack = 1;
					}else if( event == EVENT_SWITCHING0 ){
						printf( "Pausing hackindle\n" );
						system( "killall -CONT cvm" );
						usleep( 800000 );
						system( "dbus-send --system /default com.lab126.powerd.outOfScreenSaver" );
						hack = 0;
					}
				}
			}
		}
		for( i = 0; i < 5; i ++ ){
			kill( childs[i], SIGKILL );
		}
		return 0;
	}else{
		stop = 2;
		signal( SIGINT, ctrlc );
	}
	
	
	printf( "Loading background...\n" ); fflush( stdout );
	{
		int background;

		background_graphic = malloc( 800 * 600 / 2 );
		background = open( BACKGROUND_PATH, O_RDONLY );
		if( background < 0 ){
			memset( background_graphic, 0, 800 * 600 / 2 );
		}else{
			for( i = 0; i < 800 * 600 / 2; ){
				i += read( background, background_graphic + i, 800 * 600 / 2 - i );
			}
			close( background );
			for( i = 0; i < 800 * 600 / 2; i ++ ){
				background_graphic[i] = ( ( ( background_graphic[i] >> 4 ) / 4 ) << 4 ) + ( background_graphic[i] & 0xF ) / 4;
			}
		}
	}
//	printf( "Loading top bar..." ); fflush( stdout );
	{
		GifFileType *topbar[3];
		topbar[ 0 ] = DGifOpenFileName( TOPBAR_LEFT_CAP );
		topbar[ 1 ] = DGifOpenFileName( TOPBAR_MIDTEXTURE );
		topbar[ 2 ] = DGifOpenFileName( TOPBAR_RIGHT_CAP );
		eink_gifprint( background_graphic, topbar[ 0 ], 0, 0, 1 );
		eink_gifprint( background_graphic, topbar[ 1 ], topbar[ 0 ]->SWidth, 0, ( 600 - topbar[ 2 ]->SWidth - topbar[ 0 ]->SWidth ) / topbar[ 1 ]->SWidth + 1 );
		eink_gifprint( background_graphic, topbar[ 2 ], 600 - topbar[ 2 ]->SWidth, 0, 1 );
		bar_height = topbar[ 1 ]->SHeight;
		for( i = 0; i < 3; i ++ ){
			DGifCloseFile( topbar[ i ] );
		}
//		printf( " %dpx\n", bar_height );
	//	bar_height = 0;
	}
	
	FONT_HEIGHT = FONT_HEIGHT_DEFAULT;
	attr_t attr = DEFAULT_ATTR;
	
	winsz.ws_col = 600 / FONT_WIDTH;
	winsz.ws_row = ( 800 - bar_height ) / FONT_LINEHEIGHT;
	winsz.ws_xpixel = 600;
	winsz.ws_ypixel = 800 - bar_height;

	openpty( &pty, &pts, NULL, NULL, &winsz );
	
	if( ( pid = fork() ) == 0 ){
		close( pty );
		login_tty( pts );
		execlp( LOGIN_CMD, LOGIN_CMD, NULL );
		exit( 0 );
	}
	
	close( pts );
	
	printf( "Initializing eink...\n" ); fflush( stdout );
	eink_init( background_graphic, FONT_PATH, FONT_WIDTH, FONT_HEIGHT, FONT_LINEHEIGHT, bar_height );
	eink_ignore_flush( 1 );
	printf( "Initializing terminal buffer %d (%dx%d)...\n", termnum, winsz.ws_col, winsz.ws_row ); fflush( stdout );
	terminal_init( &term, MAX_BUFFER, winsz.ws_col, winsz.ws_row );
	printf( "Initializing keypad...\n" ); fflush( stdout );
	kev = kkeypad_init( pty, get_selected, &term );
	kkeypad_ignore( 1 );
//	printf( "Starting...\n" ); fflush( stdout );

	strcpy( term.title, APPLICATION_NAME );

	if( pty > kev ) fd_max = pty; else fd_max = kev;
	fd_max ++;
	
	subcursor.x = subcursor.y = 0;
	startpos.x = startpos.y = 0;
	
	int flush = 0;
	int aa = 0;
	int show_subcursor = 0;
	int show_subcursor_tmp = 0;
	unsigned long key_buf = 0;
	unsigned long key_buf_old = 0;
	
	while( 1 ){
		int status;
		if( waitpid( pid, &status, WNOHANG ) != 0 ){
			close( pty );
			openpty( &pty, &pts, NULL, NULL, &winsz );
			
			if( ( pid = fork() ) == 0 ){
				close( pty );
				login_tty( pts );
				execlp( LOGIN_CMD, LOGIN_CMD, NULL );
				exit( 0 );
			}
			close( pts );
			
			eink_reinit( background_graphic, FONT_WIDTH, FONT_HEIGHT, FONT_LINEHEIGHT, bar_height );
			terminal_reinit( &term, MAX_BUFFER, winsz.ws_col, winsz.ws_row );
			usleep( 100000 );
			ioctl( pty, TIOCSWINSZ, &winsz );

			strcpy( term.title, APPLICATION_NAME );
			subcursor.x = subcursor.y = 0;
			startpos.x = startpos.y = 0;
			attr = DEFAULT_ATTR;
			if( hack ){
				eink_clear_lines( background_graphic, 0, winsz.ws_row, attr );
				eink_flush( 1, 0 );
			}
		}

		if( flush && hack ){
			eink_flush( 7, 0 );
			flush = 0;
			aa = 1;
		}
		show_subcursor_tmp = 0;
		FD_ZERO( &key_rfds );
		FD_SET( pty, &key_rfds );
		FD_SET( kev, &key_rfds );
		timeout.tv_sec = 1;
		timeout.tv_usec = 500000;
		if( select( fd_max, &key_rfds, NULL, NULL, &timeout ) > 0 ){
			if( FD_ISSET( kev, &key_rfds ) ){
				enum EVENT event;

				read( kev, &event, sizeof( event ) );

				if( !hack ){
					if( event == EVENT_SWITCHING1 + termnum ){
						printf( "Switching to %d...\n", termnum ); fflush( stdout );
						usleep( 1000000 );
						terminal_refreash_screen( background_graphic, &term );
						eink_ignore_flush( 0 );
						eink_flush( 2, 0 );
						usleep( 100000 );
						hack = 1;
						kkeypad_ignore( 0 );
						continue;
					}
				}else if( EVENT_SWITCHING0 <= event && event <= EVENT_SWITCHING5 && event != EVENT_SWITCHING1 + termnum ){
					kkeypad_ignore( 1 );
					eink_ignore_flush( 1 );
					usleep( 1000000 );
					hack = 0;
					continue;
				}
				if( !hack ) continue;
				
				switch( event ){
				case EVENT_ROULETTE_KEY:
					read( kev, &key_buf, sizeof( key_buf ) );
					break;
				case EVENT_SUBCURSOR_UP:
					subcursor.y --;
					show_subcursor = 2;
					flush = 1;
					cur_move = 1;
					break;
				case EVENT_SUBCURSOR_DOWN:
					subcursor.y ++;
					show_subcursor = 2;
					flush = 1;
					cur_move = 1;
					break;
				case EVENT_SUBCURSOR_LEFT:
					subcursor.x --;
					show_subcursor = 2;
					flush = 1;
					cur_move = 1;
					break;
				case EVENT_SUBCURSOR_RIGHT:
					subcursor.x ++;
					show_subcursor = 2;
					flush = 1;
					cur_move = 1;
					break;
				case EVENT_SUBCURSOR_STORE:
					startpos = subcursor;
					cur_move = 1;
					break;
				case EVENT_SUBCURSOR_CLICK:
				//	show_subcursor_tmp = 2;
				//	flush = 1;
					break;
				case EVENT_HOME:
					term.ibuffer_disp = term.ibuffer_current;
					term.buffer_disp = term.buffer_current;
					terminal_refreash_screen( background_graphic, &term );
					if( hack ){
						eink_flush( 2, 0 );
						usleep( 100000 );
					}
					break;
				case EVENT_NEXT_PAGE:
					terminal_nextpage( &term );
					terminal_refreash_screen( background_graphic, &term );
					if( hack ) eink_flush( 1, 0 );
					break;
				case EVENT_PREV_PAGE:
					terminal_prevpage( &term );
					terminal_refreash_screen( background_graphic, &term );
					if( hack ) eink_flush( 1, 0 );
					break;
				case EVENT_SUBCURSOR_ON:
					show_subcursor = 1;
					break;
				case EVENT_SUBCURSOR_OFF:
					if( show_subcursor > 1 ){
						 terminal_print_selected( background_graphic, &term, 
								startpos, subcursor, 0 );
					}
					show_subcursor = 0;
					break;
				case EVENT_TEXT_LARGE:
					FONT_HEIGHT += 4;
					if( FONT_HEIGHT > 64 ) FONT_HEIGHT = 64;
					winsz.ws_col = 600 / FONT_WIDTH;
					winsz.ws_row = ( 800 - bar_height ) / FONT_LINEHEIGHT;
					eink_reinit( background_graphic, FONT_WIDTH, FONT_HEIGHT, FONT_LINEHEIGHT, bar_height );
					terminal_reinit( &term, MAX_BUFFER, winsz.ws_col, winsz.ws_row );
					usleep( 100000 );
					ioctl( pty, TIOCSWINSZ, &winsz );
					break;
				case EVENT_TEXT_SMALL:
					FONT_HEIGHT -= 4;
					if( FONT_HEIGHT < 8 ) FONT_HEIGHT = 8;
					winsz.ws_col = 600 / FONT_WIDTH;
					winsz.ws_row = ( 800 - bar_height ) / FONT_LINEHEIGHT;
					eink_reinit( background_graphic, FONT_WIDTH, FONT_HEIGHT, FONT_LINEHEIGHT, bar_height );
					terminal_reinit( &term, MAX_BUFFER, winsz.ws_col, winsz.ws_row );
					usleep( 100000 );
					ioctl( pty, TIOCSWINSZ, &winsz );
					break;
				default:
					break;
				}
				if( subcursor.x < 0 ){
					subcursor.x = winsz.ws_col - 1;
					subcursor.y --;
				}
				if( subcursor.y < 0 ) subcursor.y = 0;
				if( subcursor.x >= winsz.ws_col ){
					subcursor.x = 0;
					subcursor.y ++;
				}
				if( subcursor.y >= winsz.ws_row ){
					subcursor.y = winsz.ws_row - 1;
				}
				flush = 1;
			}
			if( FD_ISSET( pty, &key_rfds ) ){
				struct pollfd poll_sh;
				
				if( terminal_input( background_graphic, &term, pty, &attr ) < 0 ){
					continue;
				}
				flush = 1;
				
				poll_sh.fd = pty;
				poll_sh.events = POLLIN;
				if( poll( &poll_sh, 1, 0 ) ){
					flush = 0;
					continue;
				}
				usleep( 20000 );
				poll_sh.fd = pty;
				poll_sh.events = POLLIN;
				if( poll( &poll_sh, 1, 0 ) ){
					flush = 0;
					continue;
				}
			}
		}else{
			if( aa ){
				eink_font_mono( 0 );
				terminal_refreash_screen( background_graphic, &term );
				if( show_subcursor > 1 ){
					terminal_print_selected( background_graphic, &term, 
							startpos, subcursor, 1 );
				}
				eink_font_mono( 1 );
				eink_flush( 7, 0 );
				aa = 0;
			}
		}
		
		if( key_buf && key_buf_old != key_buf && term.cursor.x < winsz.ws_col &&  term.ibuffer_disp ==  term.ibuffer_current ){
			eink_putc( background_graphic, term.cursor.x, term.cursor.y, key_buf, attr );
			flush = 1;
		}
		
		key_buf_old = key_buf;
		
		if( show_subcursor > 1 ){
			terminal_print_selected( background_graphic, &term, 
					startpos, subcursor, 1 );
		}
		if( validate_pos( &term, &term.cursor ) ) terminal_print_cursor( background_graphic, &term, attr );
	}
}
