#ifndef __KINDLE_KEYPAD__
#define __KINDLE_KEYPAD__

#include "terminal.h"


#define KEYCODE_SPACE	0x39
#define KEYCODE_DELETE	0x0E
#define KEYCODE_ENTER	0x1C
#define KEYCODE_SHIFT	0x2A
#define KEYCODE_aA		0xBE
#define KEYCODE_CTRL	KEYCODE_aA
#define KEYCODE_HOME	0x66
#define KEYCODE_MENU	0x8B
#define KEYCODE_BACK	0x9E
#define KEYCODE_NEXT0	0x68
#define KEYCODE_NEXT1	0xBF
#define KEYCODE_PREV0	0x6D
#define KEYCODE_PREV1	0xC1
#define KEYCODE_ALT		0x38
#define KEYCODE_SYM		0x7E

#define KEYPAD_UP		0x67
#define KEYPAD_DOWN		0x6C
#define KEYPAD_LEFT		0x69
#define KEYPAD_RIGHT	0x6A
#define KEYPAD_CLICK	0xC2

#define VOLUME_DOWN		0x72
#define VOLUME_UP		0x73

#define KEY_CONT	2
#define KEY_DOWN	1
#define KEY_UP		0

struct input_event
{
	struct timeval time;
	unsigned short type;
	unsigned short code;
	long value;
};

typedef unsigned long *(*get_selected_func_t)(int*);

int kkeypad_init( int out, get_selected_func_t get_selected_f, struct terminal *term );
void *key_poller_therad( void *data );
void key_event( struct input_event *kbe, int stdinput, int eventinput );
void kkeypad_ignore( int state );


const static char keycode[256] =
{
	( 0 ), ( 0 ), ('1'), ('2'), ('3'), ('4'), ('5'), ('6'), ('7'), ('8'), ('9'), ('0'), ( 0 ), ( 0 ), (0x7F),( 0 ),		// 0
	('q'), ('w'), ('e'), ('r'), ('t'), ('y'), ('u'), ('i'), ('o'), ('p'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ('a'), ('s'),		// 1
	('d'), ('f'), ('g'), ('h'), ('j'), ('k'), ('l'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ('z'), ('x'), ('c'), ('v'),		// 2
	('b'), ('n'), ('m'), ( 0 ), ('.'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), (' '), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), 	// 3
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), 	// 4
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), 	// 5
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), 	// 6
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), 	// 7
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ('\t'),( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 8
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), (0x1B),( 0 ),		// 9
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// A
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// B
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// C
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// D
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// E
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// F
};
const static char shifts[256] =
{//   0      1      2      3      4      5      6      7      8      9      A      B      C      D      E      F
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ('|'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 0
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 1
	('_'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), (','), ( 0 ),		// 2
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 3
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 4
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 5
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 6
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 7
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 8
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 9
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// A
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// B
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// C
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// D
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// E
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// F
};
const static char symbols[256] =
{//   0      1      2      3      4      5      6      7      8      9      A      B      C      D      E      F
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ('|'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 0
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 1
	('_'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), (','), ( 0 ),		// 2
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 3
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 4
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 5
	( 0 ), ('~'), ('>'), ( 0 ), ( 0 ), ('#'), ( 0 ), ( 0 ), ( 0 ), ('*'), (';'), ('\''),('\"'),('/'), ('?'), ('('),		// 6
	(')'), ('!'), ('$'), ('`'), ('%'), ('&'), ('<'), ('@'), ( 0 ), ('^'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 7
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 8
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 9
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// A
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// B
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// C
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// D
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// E
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// F
};
const static char alts[256] =
{//   0      1      2      3      4      5      6      7      8      9      A      B      C      D      E      F
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 0
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 1
	('_'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ('\\'),( 0 ),		// 2
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 3
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 4
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 5
	( 0 ), ('~'), ('>'), ( 0 ), ( 0 ), ('#'), ( 0 ), ( 0 ), (':'), ('*'), (';'), ('\''),('\"'),('/'), ('?'), ('('),		// 6
	(')'), ('!'), ('$'), ('`'), ('%'), ('&'), ('<'), ('@'), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 7
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 8
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// 9
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// A
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// B
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// C
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// D
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// E
	( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ), ( 0 ),		// F
};
const static char alt_menu[] = { '-', '=', '+', 0 };
const static char alt_del[] = { '{', '}', '[', ']', 0 };


enum EVENT {
	EVENT_SPECIAL_CHARACTOR,
	EVENT_SUBCURSOR_UP,
	EVENT_SUBCURSOR_DOWN,
	EVENT_SUBCURSOR_LEFT,
	EVENT_SUBCURSOR_RIGHT,
	EVENT_SUBCURSOR_STORE,
	EVENT_SUBCURSOR_CLICK,
	EVENT_SWITCHING0,
	EVENT_SWITCHING1,
	EVENT_SWITCHING2,
	EVENT_SWITCHING3,
	EVENT_SWITCHING4,
	EVENT_SWITCHING5,
	EVENT_HOME,
	EVENT_NEXT_PAGE,
	EVENT_PREV_PAGE,
	EVENT_TEXT_LARGE,
	EVENT_TEXT_SMALL,
	EVENT_SUBCURSOR_ON,
	EVENT_SUBCURSOR_OFF,
	EVENT_ROULETTE_KEY,
	EVENT_MAX
};

struct EVENT_KEYSET {
	enum EVENT event;
	char *str;
	unsigned long code;
	int type;
	int pressed[2];
	int unpressed;
	
};

const static struct EVENT_KEYSET event_key[] = {
	{ EVENT_SPECIAL_CHARACTOR,	"\033 ",	KEYCODE_SPACE,	KEY_DOWN,	{KEYCODE_CTRL,0},	KEYCODE_SHIFT	},
	{ EVENT_SPECIAL_CHARACTOR,	" ",		KEYCODE_SPACE,	KEY_DOWN,	{0}, 				KEYCODE_SHIFT	},
	{ EVENT_SPECIAL_CHARACTOR,	"\x09",		KEYCODE_MENU,	KEY_DOWN,	{0}, 				KEYCODE_SHIFT	},
	{ EVENT_SPECIAL_CHARACTOR,	"\x0D",		KEYCODE_ENTER,	KEY_DOWN,	{0}, 				KEYCODE_SHIFT	},
	{ EVENT_SPECIAL_CHARACTOR,	"\033[6~",	KEYCODE_NEXT0,	KEY_DOWN,	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SPECIAL_CHARACTOR,	"\033[6~",	KEYCODE_NEXT1,	KEY_DOWN,	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SPECIAL_CHARACTOR,	"\033[5~",	KEYCODE_PREV0,	KEY_DOWN,	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SPECIAL_CHARACTOR,	"\033[5~",	KEYCODE_PREV1,	KEY_DOWN,	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_TEXT_LARGE,			NULL,		KEYCODE_NEXT0,	KEY_DOWN,	{KEYCODE_HOME,0}, 	0 				},
	{ EVENT_TEXT_LARGE,			NULL,		KEYCODE_NEXT1,	KEY_DOWN,	{KEYCODE_HOME,0}, 	0 				},
	{ EVENT_TEXT_SMALL,			NULL,		KEYCODE_PREV0,	KEY_DOWN,	{KEYCODE_HOME,0}, 	0 				},
	{ EVENT_TEXT_SMALL,			NULL,		KEYCODE_PREV1,	KEY_DOWN,	{KEYCODE_HOME,0}, 	0 				},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033[A",	KEYPAD_UP, 		KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033[A",	KEYPAD_UP, 		KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033[B",	KEYPAD_DOWN, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033[B",	KEYPAD_DOWN, 	KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033[D",	KEYPAD_LEFT, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033[D",	KEYPAD_LEFT, 	KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033[C",	KEYPAD_RIGHT, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033[C",	KEYPAD_RIGHT, 	KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SUBCURSOR_UP, 		NULL,		KEYPAD_UP, 		KEY_DOWN, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_UP, 		NULL,		KEYPAD_UP, 		KEY_CONT, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_DOWN, 	NULL,		KEYPAD_DOWN, 	KEY_DOWN, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_DOWN, 	NULL,		KEYPAD_DOWN, 	KEY_CONT, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_LEFT, 	NULL,		KEYPAD_LEFT, 	KEY_DOWN, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_LEFT, 	NULL,		KEYPAD_LEFT, 	KEY_CONT, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_RIGHT, 	NULL,		KEYPAD_RIGHT, 	KEY_DOWN, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_RIGHT, 	NULL,		KEYPAD_RIGHT, 	KEY_CONT, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_STORE, 	NULL,		KEYPAD_CLICK, 	KEY_UP, 	{KEYCODE_SHIFT,0}, 	0 				},
	{ EVENT_SUBCURSOR_CLICK, 	NULL,		KEYPAD_CLICK, 	KEY_UP, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SWITCHING0,			NULL,		0x10,			KEY_DOWN,	{KEYCODE_SHIFT,
																		 KEYCODE_aA},		0 				},
	{ EVENT_SWITCHING1,			NULL,		0x11,			KEY_DOWN,	{KEYCODE_SHIFT,
																		 KEYCODE_aA},		0 				},
	{ EVENT_SWITCHING2,			NULL,		0x12,			KEY_DOWN,	{KEYCODE_SHIFT,
																		 KEYCODE_aA},		0 				},
	{ EVENT_SWITCHING3,			NULL,		0x13,			KEY_DOWN,	{KEYCODE_SHIFT,
																		 KEYCODE_aA},		0 				},
	{ EVENT_SWITCHING4,			NULL,		0x14,			KEY_DOWN,	{KEYCODE_SHIFT,
																		 KEYCODE_aA},		0 				},
	{ EVENT_SWITCHING5,			NULL,		0x15,			KEY_DOWN,	{KEYCODE_SHIFT,
																		 KEYCODE_aA},		0 				},
	{ EVENT_HOME,				NULL,		KEYCODE_HOME,	KEY_UP,		{0},		 		KEYCODE_SHIFT	},
	{ EVENT_NEXT_PAGE,			NULL,		KEYCODE_NEXT0,	KEY_UP,		{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_NEXT_PAGE,			NULL,		KEYCODE_NEXT1,	KEY_UP,		{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_PREV_PAGE,			NULL,		KEYCODE_PREV0,	KEY_UP,		{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_PREV_PAGE,			NULL,		KEYCODE_PREV1,	KEY_UP,		{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SUBCURSOR_ON,		NULL,		KEYCODE_SHIFT,	KEY_DOWN,	{0}, 				0				},
	{ EVENT_SUBCURSOR_OFF,		NULL,		KEYCODE_SHIFT,	KEY_UP,		{0}, 				0				},
	{ EVENT_MAX,				NULL,		0,				0,			{0}, 				0			 	},
};

const static struct EVENT_KEYSET appcur_event_key[] = {
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OA",	KEYPAD_UP, 		KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OA",	KEYPAD_UP, 		KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OB",	KEYPAD_DOWN, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OB",	KEYPAD_DOWN, 	KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OD",	KEYPAD_LEFT, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OD",	KEYPAD_LEFT, 	KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OC",	KEYPAD_RIGHT, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OC",	KEYPAD_RIGHT, 	KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_MAX,				NULL,		0,				0,			{0}, 				0			 	},
};

const static struct EVENT_KEYSET appkey_event_key[] = {
	{ EVENT_SPECIAL_CHARACTOR,	"\000",		KEYCODE_SPACE,	KEY_DOWN,	{KEYCODE_CTRL,0},	KEYCODE_SHIFT	},
/*	{ EVENT_SPECIAL_CHARACTOR, 	"\033O ",	KEYCODE_SPACE, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033O ",	KEYCODE_SPACE, 	KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OI",	KEYCODE_MENU, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OM",	KEYCODE_ENTER, 	KEY_DOWN, 	{0}, 				KEYCODE_SHIFT 	},
	{ EVENT_SPECIAL_CHARACTOR, 	"\033OM",	KEYCODE_ENTER, 	KEY_CONT, 	{0}, 				KEYCODE_SHIFT 	},*/
	{ EVENT_MAX,				NULL,		0,				0,			{0}, 				0			 	},
};
#endif
