#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <unistd.h>
#include <pty.h>
#include <utmp.h>
#include <sys/wait.h>

#include "terminal.h"
#include "eink.h"

//#define TERMINAL_DEBUG

void attr_extract( attr_t attr, unsigned char *fg, unsigned char *bg, unsigned char *under )
{
	*fg = ATTR_FG( attr );
	*bg = ATTR_BG( attr );
	
	if( attr & ATTR_INVERSE ){
		unsigned char tmp;
		tmp = *fg;
		*fg = *bg;
		*bg = tmp;
	}
	if( attr & ATTR_UNDERLINE ){
		*under = *fg;
	}else{
		*under = *bg;
	}
}


int validate_pos( struct terminal *term, struct position *pos )
{
	if( pos->x < 0 || pos->y < 0 ) return 0;
	if( pos->x >= term->buffer_current->column ) return 0;
	if( pos->y >= term->buffer_current->row ) return 0;
	return 1;
}

void terminal_clear( struct terminal *term, int ibuffer, int xfrom, int yfrom, int xto, int yto )
{
	if( xto >= term->buffer[ ibuffer ].column ) xto = term->buffer[ ibuffer ].column;
	if( yto > term->buffer[ ibuffer ].row ){
		yto = term->buffer[ ibuffer ].row;
		xto = term->buffer[ ibuffer ].column;
	}
	if( xto - xfrom + ( yto - yfrom ) * term->buffer[ ibuffer ].column < 0 ) return;
	
	memset( &term->buffer[ ibuffer ].text[ xfrom + yfrom * term->buffer[ ibuffer ].column ], 0,
									( xto - xfrom + ( yto - yfrom ) * term->buffer[ ibuffer ].column ) * sizeof( unsigned long ) );
	memset( &term->buffer[ ibuffer ].attr[ xfrom + yfrom * term->buffer[ ibuffer ].column ], 0,
									( xto - xfrom + ( yto - yfrom ) * term->buffer[ ibuffer ].column ) * sizeof( attr_t ) );

}

void terminal_delete_line( struct terminal *term, int ibuffer, int row, int num )
{
	int i;
//	printf("%d,%d\n",row,num);
	for( i = row; i <= term->buffer[ ibuffer ].scroll_bottom; i ++ ){
		if( i + num <= term->buffer[ ibuffer ].scroll_bottom ){
			memcpy( &term->buffer[ ibuffer ].text[ i * term->buffer[ ibuffer ].column ], 
											&term->buffer[ ibuffer ].text[ ( i + num ) * term->buffer[ ibuffer ].column ],
											term->buffer[ ibuffer ].column * sizeof( unsigned long ) );
			memcpy( &term->buffer[ ibuffer ].attr[ i * term->buffer[ ibuffer ].column ], 
											&term->buffer[ ibuffer ].attr[ ( i + num ) * term->buffer[ ibuffer ].column ],
											term->buffer[ ibuffer ].column * sizeof( attr_t ) );
		}else{
			memset( &term->buffer[ ibuffer ].text[ i * term->buffer[ ibuffer ].column ], 
											0, term->buffer[ ibuffer ].column * sizeof( unsigned long ) );
			memset( &term->buffer[ ibuffer ].attr[ i * term->buffer[ ibuffer ].column ], 
											0, term->buffer[ ibuffer ].column * sizeof( attr_t ) );
		}
	}
}

void terminal_insert_line( struct terminal *term, int ibuffer, int row, int num )
{
	int i;
	for( i = term->buffer[ ibuffer ].scroll_bottom; i >= row; i -- ){
		if( i - num >= row ){
			memcpy( &term->buffer[ ibuffer ].text[ i * term->buffer[ ibuffer ].column ], 
											&term->buffer[ ibuffer ].text[ ( i - num ) * term->buffer[ ibuffer ].column ],
											term->buffer[ ibuffer ].column * sizeof( unsigned long ) );
			memcpy( &term->buffer[ ibuffer ].attr[ i * term->buffer[ ibuffer ].column ], 
											&term->buffer[ ibuffer ].attr[ ( i - num ) * term->buffer[ ibuffer ].column ],
											term->buffer[ ibuffer ].column * sizeof( attr_t ) );
		}else{
			memset( &term->buffer[ ibuffer ].text[ i * term->buffer[ ibuffer ].column ], 
											0, term->buffer[ ibuffer ].column * sizeof( unsigned long ) );
			memset( &term->buffer[ ibuffer ].attr[ i * term->buffer[ ibuffer ].column ], 
											0, term->buffer[ ibuffer ].column * sizeof( attr_t ) );
		}
	}
}

void terminal_delete_char( struct terminal *term, int ibuffer, int row, int col, int num )
{
	memmove( &term->buffer[ ibuffer ].text[ row * term->buffer[ ibuffer ].column + col ], 
									&term->buffer[ ibuffer ].text[ row * term->buffer[ ibuffer ].column + col + num ],
									( term->buffer[ ibuffer ].column - col - num ) * sizeof( unsigned long ) );
	memmove( &term->buffer[ ibuffer ].attr[ row * term->buffer[ ibuffer ].column + col ], 
									&term->buffer[ ibuffer ].attr[ row * term->buffer[ ibuffer ].column + col + num ],
									( term->buffer[ ibuffer ].column - col - num ) * sizeof( attr_t ) );
	memset( &term->buffer[ ibuffer ].text[ row * term->buffer[ ibuffer ].column + term->buffer[ ibuffer ].column - num ], 
									0, num * sizeof( unsigned long ) );
	memset( &term->buffer[ ibuffer ].attr[ row * term->buffer[ ibuffer ].column + term->buffer[ ibuffer ].column - num ], 
									0, num * sizeof( attr_t ) );
}

void terminal_insert_char( struct terminal *term, int ibuffer, int row, int col, int num )
{
	memmove( &term->buffer[ ibuffer ].text[ row * term->buffer[ ibuffer ].column + col + num ], 
									&term->buffer[ ibuffer ].text[ row * term->buffer[ ibuffer ].column + col ],
									( term->buffer[ ibuffer ].column - col - num ) * sizeof( unsigned long ) );
	memmove( &term->buffer[ ibuffer ].attr[ row * term->buffer[ ibuffer ].column + col + num ], 
									&term->buffer[ ibuffer ].attr[ row * term->buffer[ ibuffer ].column + col ],
									( term->buffer[ ibuffer ].column - col - num ) * sizeof( attr_t ) );
	memset( &term->buffer[ ibuffer ].text[ row * term->buffer[ ibuffer ].column + col ], 
									0, num * sizeof( unsigned long ) );
	memset( &term->buffer[ ibuffer ].attr[ row * term->buffer[ ibuffer ].column + col ], 
									0, num * sizeof( attr_t ) );
}

unsigned long terminal_getchar( struct terminal *term, int ibuffer, int x, int y )
{
	return term->buffer[ ibuffer ].text[ x + y * term->buffer[ ibuffer ].column ];
}

unsigned long terminal_getchar_cursor( struct terminal *term )
{
	return term->buffer_current->text[ term->cursor.x + term->cursor.y * term->buffer_current->column ];
}

attr_t terminal_getattr( struct terminal *term, int ibuffer, int x, int y )
{
	return term->buffer[ ibuffer ].attr[ x + y * term->buffer[ ibuffer ].column ];
}

void terminal_setchar( struct terminal *term, int ibuffer, int x, int y, unsigned long ch, attr_t attr )
{
	term->buffer[ ibuffer ].text[ x + y * term->buffer[ ibuffer ].column ] = ch;
	term->buffer[ ibuffer ].attr[ x + y * term->buffer[ ibuffer ].column ] = attr;
}

void terminal_reset( struct terminal *term )
{
	int i;
	
	memset( term->decpm, 0, sizeof( term->decpm ) );
	term->decpm[DECPM_ShowCursor] = 1;
	term->decpm[DECPM_WraparoundMode] = 1;
	term->decpm[DECPM_ApplicationKeypad] = 0;
	memset( term->setmode, 0, sizeof( term->setmode ) );

	term->ibuffer_disp = term->ibuffer_current = 0;
	term->buffer_current = &term->buffer[ term->ibuffer_current ];
	term->buffer_disp = &term->buffer[ term->ibuffer_disp ];

	for( i = 0; i < term->nbuffer + 1; i ++ ){
		terminal_clear( term, i, 0, 0, term->buffer[ i ].column, term->buffer[ i ].row );
	}
	term->cursor.x = term->cursor.y = 0;
}

void terminal_init( struct terminal *term, int nbuffer, int width, int height )
{
	int i;
	
	term->nbuffer = nbuffer;
	term->buffer = malloc( ( nbuffer + 1 ) * sizeof( struct terminal_buffer ) );
//	printf( "Init - 0x%x\n", term->buffer );

	for( i = 0; i < nbuffer + 1; i ++ ){
		term->buffer[ i ].text = malloc( ( width * (height + 1) + 1 ) * sizeof( unsigned long ) );
		term->buffer[ i ].attr = malloc( ( width * (height + 1) + 1 ) * sizeof( attr_t ) );
//		printf( " - 0x%x 0x%x\n", term->buffer[ i ].text, term->buffer[ i ].attr );
		term->buffer[ i ].column = width;
		term->buffer[ i ].row = height;
		term->buffer[ i ].scroll_top = 0;
		term->buffer[ i ].scroll_bottom = height - 1;
	}
	terminal_reset( term );
}

void terminal_free( struct terminal *term )
{
	int i;
//	printf( "free - 0x%x\n", term->buffer );
	for( i = term->nbuffer; i > 0; i -- ){
//		printf( " - 0x%x 0x%x\n", term->buffer[ i ].text, term->buffer[ i ].attr );
		free( ( term->buffer[ i ].text ) );
		free( ( term->buffer[ i ].attr ) );
	}
	free( term->buffer );
}

void terminal_reinit( struct terminal *term, int nbuffer, int width, int height )
{
	struct terminal term2;
	struct terminal term_new;
	
	term2 = *term;
	
	terminal_init( &term_new, nbuffer, width, height );
	*term = term_new;
	
	strcpy( term->title, term2.title );
	
	terminal_free( &term2 );
}

void terminal_nextpage( struct terminal *term )
{
	if( term->ibuffer_current == term->ibuffer_disp || term->decpm[ DECPM_UseAlternateScreenBuffer ] || 
			term->ibuffer_current >= term->nbuffer ){
		return;
	}
	term->ibuffer_disp ++;
	term->ibuffer_disp %= term->nbuffer;
	term->buffer_disp = &term->buffer[ term->ibuffer_disp ];
}

void terminal_prevpage( struct terminal *term )
{
	if( term->ibuffer_current == ( term->ibuffer_disp - 1 + term->nbuffer ) % term->nbuffer || 
			term->decpm[ DECPM_UseAlternateScreenBuffer ] || 
			term->ibuffer_current >= term->nbuffer ){
		return;
	}
	term->ibuffer_disp --;
	term->ibuffer_disp += term->nbuffer;
	term->ibuffer_disp %= term->nbuffer;
	term->buffer_disp = &term->buffer[ term->ibuffer_disp ];
}

void terminal_print_title( char *background_graphic, struct terminal *term )
{
	int x;
	char *val;
	x = 0;
	eink_font_mono( 0 );
	eink_clear_lines( background_graphic, -1, 1, DEFAULT_ATTR );
	for( val = term->title; *val; ){
		unsigned long code;
		unsigned long key;
		int nbytes;
		key = *val;
		val ++;
		if( ( key & 0xC0 ) == 0xC0 ){
			// UTF-8
			if( !( key & 0x20 ) ){
				code = key & 0x1F;
				nbytes = 1;
			}else if( !( key & 0x10 ) ){
				code = key & 0x0F;
				nbytes = 2;
			}else if( !( key & 0x08 ) ){
				code = key & 0x07;
				nbytes = 3;
			}else if( !( key & 0x04 ) ){
				code = key & 0x03;
				nbytes = 4;
			}else if( !( key & 0x02 ) ){
				code = key & 0x01;
				nbytes = 5;
			}else{
				continue;
			}

			for( ; nbytes > 0 && *val; nbytes -- ){
				code = code << 6;
				code |= (*val) & 0x3F;
				val ++;
			}
			key = code;
		}
	//	printf("%x ", key);fflush(stdout);
		eink_putc2( background_graphic, 15 + x * 11, 7, 22, 22, 22+22/4, key, DEFAULT_ATTR );
		
		x ++;
		if( key > 0xFF ) x ++;
	}
	eink_font_mono( 1 );
}

int terminal_input( char *background_graphic, struct terminal *term, int pty, attr_t *attr )
{
	static char *stdinbuf = NULL;
	static int stdinbuf_left = 0;
	static struct position saved_cursor = { 0, 0 };
	static int ibuffer_current_store = 0;
//	int l;
	int flush = 0;
	int len;
	char *stdinpos;
	
	if( !stdinbuf ){
		stdinbuf = malloc( INBUFFER_SIZE );
		memset( stdinbuf, 0, INBUFFER_SIZE );
	}
	
	len = read( pty, stdinbuf + stdinbuf_left, INBUFFER_SIZE - 1 - stdinbuf_left );
	if( len <= 0 ){
		return -1;
	}
	len += stdinbuf_left;
	stdinbuf_left = 0;
	{
		int i, j;
		for( j = i = 0; i < len; i ++ ){
			if( stdinbuf[ i ] == 0 ){
				continue;
			}
			stdinbuf[ j ] = stdinbuf[ i ];
			j++;
		}
		len = j;
	}
	stdinpos = stdinbuf;
	stdinbuf[ len ] = 0;

	for( ; stdinpos < stdinbuf + len; ){
		unsigned long key;
		
		key = *stdinpos;
		stdinpos ++;
#ifdef TERMINAL_DEBUG
		printf("(%x) %c \n", key, (key >= 0x20 && key<0x80)?key:' ' ); fflush( stdout );
#endif

		if( key == '\r' ){
			term->cursor.x = 0;
			continue;
		}
		if( key == '\n' ){
			if( term->buffer_current->scroll_bottom + 1 != term->buffer_current->row || term->buffer_current->scroll_top != 0 ){
				if( term->cursor.y >= term->buffer_current->scroll_bottom ){
					terminal_delete_line( term,  term->ibuffer_current, term->buffer_current->scroll_top, term->cursor.y + 1 - term->buffer_current->scroll_bottom );
					terminal_refreash_screen( background_graphic, term );
					
					term->cursor.y = term->buffer_current->scroll_bottom;
					//break;
				}else{
					term->cursor.y += 1;
				}
			}else{
				term->cursor.y += 1;
			}
			continue;
		}
		if( key == '\b' ){
#ifdef TERMINAL_DEBUG
			printf(" \\b\n" ); fflush( stdout );
#endif
			term->cursor.x --;
			if( term->cursor.x < 0 ) term->cursor.x = 0;
		}
		if( key == '\t' ){
			term->cursor.x = ( term->cursor.x / 8 + 1 ) * 8;
		}
		if( key == 0x7F ){
#ifdef TERMINAL_DEBUG
			printf(" Del\n" ); fflush( stdout );
#endif
			if( validate_pos( term, &term->cursor ) ){
				eink_clear_chars( background_graphic, term->cursor.x, term->cursor.y, 1, *attr );
				terminal_setchar( term, term->ibuffer_current, term->cursor.x, term->cursor.y, 0, *attr );
			}
			term->cursor.x --;
			if( term->cursor.x < 0 ){
				term->cursor.y --;
				if( term->cursor.y < 0 ) term->cursor.y = 0;
			}
		}
		if( ( key & 0xC0 ) == 0xC0 ){
			unsigned long code;
			int nbytes;
			// UTF-8
			stdinbuf[ stdinbuf_left ] = key;
			stdinbuf_left ++;

			if( !( key & 0x20 ) ){
				code = key & 0x1F;
				nbytes = 1;
			}else if( !( key & 0x10 ) ){
				code = key & 0x0F;
				nbytes = 2;
			}else if( !( key & 0x08 ) ){
				code = key & 0x07;
				nbytes = 3;
			}else if( !( key & 0x04 ) ){
				code = key & 0x03;
				nbytes = 4;
			}else if( !( key & 0x02 ) ){
				code = key & 0x01;
				nbytes = 5;
			}else{
				continue;
			}

			for( ; nbytes > 0 && *stdinpos; nbytes -- ){
				stdinbuf[ stdinbuf_left ] = *stdinpos;
				stdinbuf_left ++;
				code = code << 6;
				code |= (*stdinpos) & 0x3F;
				stdinpos ++;
			}
			if( nbytes > 0 ){
#ifdef TERMINAL_DEBUG
				printf( "UTF8 code continues (reamains %d)\n",nbytes);
#endif
				break;
			}
			key = code;
#ifdef TERMINAL_DEBUG
			stdinbuf[ stdinbuf_left ] = 0;
			printf("(%x) %s \n", key, stdinbuf ); fflush( stdout );
#endif
			stdinbuf_left = 0;
		//	printf( "%x  %x\n",key, code );
		}
		if( key == '\033' ){
			char cmd = 0;
			char type;
			int arg[6];
//			int i;
			int narg;
			char ret[32];
			int qc;
			
			stdinbuf[ stdinbuf_left ] = key;
			stdinbuf_left ++;
			
			type = *stdinpos;
			stdinpos ++;
			if( type == 0 ){
				break;
			}
			stdinbuf[ stdinbuf_left ] = type;
			stdinbuf_left ++;
			if( type == '(' || type == ')' || type == '*' || type == '+' || type == '-' || type == '.' ){
				unsigned char char_code;
				// Character code
				char_code = *stdinpos;
				stdinpos ++;
				stdinbuf[ stdinbuf_left ] = char_code;
				stdinbuf_left ++;
				if( char_code == 0 ){
					break;
				}
				stdinbuf_left = 0;
#ifdef TERMINAL_DEBUG
				printf(" (%x) %c    %x\n", type, type, char_code ); fflush( stdout );
#endif
				continue;
			}
			if( type == ']' ){
				int i;
				char value[ TITLE_LENGTH ];
				
				cmd = *stdinpos;
				stdinpos ++;
				stdinbuf[ stdinbuf_left ] = cmd;
				stdinbuf_left ++;
				if( cmd == 0 ){
					break;
				}
				i = 0;
#ifdef TERMINAL_DEBUG
				printf("] (%x) %c    (%x) %c    \n", type, type >= 0x20?type:' ' , cmd, cmd >= 0x20?cmd:' ' ); fflush( stdout );
#endif
				for( ;; ){
					value[ i ] = *stdinpos;
					stdinpos ++;
					stdinbuf[ stdinbuf_left ] = value[ i ];
					stdinbuf_left ++;
					if( value[ i ] == 0 ){
						i = -1;
						break;
					}
					if( value[ i ] == '\a' || ( value[ i ] == '\033' && *stdinpos == '\\' ) ){
						if( value[ i ] == '\033' ) stdinpos ++;
						break;
					}
					i ++;
					if( i >= TITLE_LENGTH ) i = TITLE_LENGTH - 1;
				}
				if( i == -1 ) break;
				value[ i ] = 0;
				stdinbuf_left = 0;
#ifdef TERMINAL_DEBUG
				printf( "System Command %c %s(%d)\n", cmd, value, strlen(value) );
#endif
				switch( cmd ){
				case '0':
				case '3':
					strcpy( term->title, value + 1 );
					terminal_print_title( background_graphic, term );
					break;
				}
				continue;
			}
			if( type != '[' ){
#ifdef TERMINAL_DEBUG
				printf("![ (%x) %c    (%x) %c    \n", type, type >= 0x20?type:' ' , *stdinpos, *stdinpos >= 0x20?*stdinpos:' ' ); fflush( stdout );
#endif
			}
			switch( type ){
			case 'c':
				// Reset Device
				terminal_reset( term );
				*attr = DEFAULT_ATTR;
				break;
			case '7':
				// Save Cursor
				saved_cursor = term->cursor;
				break;
			case '8': // Restore Cursor Position
				term->cursor = saved_cursor;
				break;
			case '=': // Application Keypad (DECPAM)
				term->decpm[ DECPM_ApplicationKeypad ] = 1;
				break;
			case '>': // Normal Keypad (DECPNM)
				term->decpm[ DECPM_ApplicationKeypad ] = 0;
				break;
			}
			if( type != '[' ){
				//if( type == '\033' || type >= 0x80 ){
				//	stdinpos --;
				//}
				stdinbuf_left = 0;
				continue;
			}
			{
				int num;
				int i;
				char minus;
				char code = 0;
				narg = 0;
				arg[0] = arg[1] = arg[2] = arg[3] = arg[4] = arg[5] = -1;
				num = 0;
				i = 0;
				minus = 1;
				qc = ' ';
				for( ; *stdinpos != 0; i ++ ){
					code = *stdinpos;
					stdinpos ++;
					stdinbuf[ stdinbuf_left ] = code;
					stdinbuf_left ++;
					if( ( code == '?' || code == '>' ) && i == 0 ){
						qc = code;
						continue;
					}
					if( code == '-' ){
						minus = -1;
					}else if( code < '0' || '9' < code ){
						cmd = code;
						arg[ narg ] = num * minus;
						num = 0;
						narg ++; if( narg > 5 ) narg = 5;
						if( cmd != ';' ){
							break;
						}
					}else{
						num *= 10;
						num += code - '0';
					}
				}
				if( code == 0 ){
					break;
				}
				stdinbuf_left = 0;
			}
			if( qc == '?' ){
				if( cmd == 'l' ){
					if( arg[ 0 ] < DECPM_MAX ) term->decpm[arg[0]] = 0;
#ifdef TERMINAL_DEBUG
					printf(" %d  Off\n", arg[0] ); fflush( stdout );
#endif
					switch( arg[ 0 ] ){
					case DECPM_SaveCursorAndUseAlternateScreenBuffer:
						term->cursor = saved_cursor;
					case DECPM_UseAlternateScreenBuffer:
						term->ibuffer_disp = term->ibuffer_current = ibuffer_current_store;
						term->buffer_current =  &term->buffer[term->ibuffer_current];
						term->buffer_disp = &term->buffer[ term->ibuffer_disp ];
						terminal_refreash_screen( background_graphic, term );
						eink_flush( 2, 0 );
						break;
					}
				}else if( cmd == 'h' ){
					if( arg[ 0 ] < DECPM_MAX ) term->decpm[arg[0]] = 1;
#ifdef TERMINAL_DEBUG
					printf(" %d  ON\n", arg[0] ); fflush( stdout );
#endif
					switch( arg[ 0 ] ){
					case DECPM_SaveCursorAndUseAlternateScreenBuffer:
						saved_cursor = term->cursor;
					case DECPM_UseAlternateScreenBuffer:
						ibuffer_current_store = term->ibuffer_current;
						term->ibuffer_current = term->ibuffer_disp = term->nbuffer;
						term->buffer_current =  &term->buffer[term->ibuffer_disp];
						term->buffer_disp = &term->buffer[ term->ibuffer_disp ];
						if( arg[ 0 ] == DECPM_SaveCursorAndUseAlternateScreenBuffer ){
							terminal_clear( term, term->ibuffer_current, 0, 0, term->buffer_current->column, term->buffer_current->row );
							term->cursor.x = term->cursor.y = 0;
						}
						terminal_refreash_screen( background_graphic, term );
						eink_flush( 2, 0 );
						break;
					}
				}else{
#ifdef TERMINAL_DEBUG
					printf("? %c %d  ON\n", cmd, arg[0] ); fflush( stdout );
#endif
				}
				continue;
			}
			if( qc != ' ' ){
			//	printf(" %c %c  (%x),%d,%d,%d  %d,%d\n", qc, cmd, cmd, arg[0], arg[1], arg[2], term->buffer_current->column, term->buffer_current->row ); fflush( stdout );
			}
#ifdef TERMINAL_DEBUG
			printf(" (%x)%c  ,%d,%d,%d,%d,%d          (%x)%c          \n", cmd, cmd >= 0x20?cmd:' ', arg[0], arg[1], arg[2], arg[3], arg[4], *stdinpos, *stdinpos >= 0x20?*stdinpos:' ' ); fflush( stdout );
#endif
			switch( cmd )
			{
			//////////////////////
			// ANSI VT100
			case 'A': // Cursor Up
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.y -= arg[0];
				if( term->cursor.y < 0 ) term->cursor.y = 0;
				break;
			case 'B': // Cursor Down
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.y += arg[0];
				if( term->cursor.y >= term->buffer_current->row ) term->cursor.y = term->buffer_current->row - 1;
				break;
			case 'C': // Cursor Forward
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.x += arg[0];
				if( term->cursor.x >= term->buffer_current->column ){
					term->cursor.x = 0;
					term->cursor.y += 1;
					if( term->cursor.y >= term->buffer_current->row ) term->cursor.y = term->buffer_current->row - 1;
				}
				break;
			case 'D': // Cursor Back
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.x -= arg[0];
				if( term->cursor.x < 0 ){
					term->cursor.x = 0;
					term->cursor.y -= 1;
					if( term->cursor.y < 0 ) term->cursor.y = 0;
				}
				break;
			case 'E': // Cursor Next Line
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.x = 0;
				term->cursor.y += arg[0];
				break;
			case 'F': // Cursor Previous Line
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.x = 0;
				term->cursor.y -= arg[0];
				break;
			case 'G': // Cursor Horizontal Absolute
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.x = arg[0] - 1;
				break;
			case 'H': // Cursor Position
			case 'f': 
				if( term->ibuffer_disp != term->ibuffer_current ){
					term->ibuffer_disp = term->ibuffer_current;
					eink_clear_lines( background_graphic, 0, term->buffer_current->row, DEFAULT_ATTR );
					terminal_refreash_screen( background_graphic, term );
					eink_flush( 2, 0 );
				}
				if( arg[0] <= 0 ) arg[0] = 1;
				if( arg[1] <= 0 ) arg[1] = 1;
				

				if( term->cursor.y > arg[0] - 1 ){
					eink_flush( 1, 0 );
				}
				
				term->cursor.y = arg[0] - 1;
				term->cursor.x = arg[1] - 1;
				if( term->cursor.x >= term->buffer_current->column ){
					if( term->decpm[ DECPM_WraparoundMode ] && !term->decpm[ DECPM_UseAlternateScreenBuffer ] ){
						term->cursor.x = 0;
						term->cursor.y ++;
					}else{
						term->cursor.x = term->buffer_current->column - 1;
					}
				}
				break;
			case 'd': // Line Position Absolute
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.y = arg[0] - 1;
				break;
			case '`': // Character Position Absolute 
				if( arg[0] <= 0 ) arg[0] = 1;
				term->cursor.x = arg[0] - 1;
				break;
			case 'h':
				if( arg[0] < 0 ) break;
				term->setmode[ arg[0] ] = 1;
				break;
			case 'l':
				if( arg[0] < 0 ) break;
				term->setmode[ arg[0] ] = 0;
				break;
			case 'L': // Insert Lines
				if( arg[0] <= 0 ) arg[0] = 1;
				terminal_insert_line( term,  term->ibuffer_current, term->cursor.y, arg[ 0 ] );
				terminal_refreash_screen( background_graphic, term );
				break;
			case 'M': // Delete Lines
				if( arg[0] <= 0 ) arg[0] = 1;
				terminal_delete_line( term,  term->ibuffer_current, term->cursor.y, arg[ 0 ] );
				terminal_refreash_screen( background_graphic, term );
				break;
			case 'P': // Delete Characters
				// TODO: 前の行
				if( arg[0] <= 0 ) arg[0] = 1;
				terminal_delete_char( term,  term->ibuffer_current, term->cursor.y, term->cursor.x, arg[ 0 ] );
				terminal_refreash_line( background_graphic, term, term->cursor.y );
				break;
			case '@': // Insert blank charactor
				if( arg[0] <= 0 ) arg[0] = 1;
				terminal_insert_char( term,  term->ibuffer_current, term->cursor.y, term->cursor.x, arg[ 0 ] );
				terminal_refreash_line( background_graphic, term, term->cursor.y );
				break;
			case 'J': // Erase Data
				if( arg[0] == -1 ) arg[0] = 0;
				switch( arg[0] ){
				case 0: // Clear from cursor to end of screen
					if( term->ibuffer_disp == term->ibuffer_current ){
						if( term->cursor.y < term->buffer_current->row )
							eink_clear_chars( background_graphic, term->cursor.x, term->cursor.y, 
											term->buffer_current->column - term->cursor.x - 1, DEFAULT_ATTR );
						eink_clear_lines( background_graphic, 
											term->cursor.y + 1, term->buffer_current->row - term->cursor.y - 1, DEFAULT_ATTR );
					}
					terminal_clear( term, term->ibuffer_current, term->cursor.x, term->cursor.y, 
											term->buffer[ term->ibuffer_current ].column, term->buffer[ term->ibuffer_current ].row );
					break;
				case 1: // Clear from cursor to beginning of screen
					if( term->ibuffer_disp == term->ibuffer_current ){
						if( term->cursor.y < term->buffer_current->row )
							eink_clear_chars( background_graphic, 0, term->cursor.y, term->cursor.x - 1, DEFAULT_ATTR );
						eink_clear_lines( background_graphic, 0, term->cursor.y - 1, DEFAULT_ATTR );
					}
					terminal_clear( term, term->ibuffer_current, 0, 0, term->cursor.x, term->cursor.y );
					break;
				case 2: // Clear entire screen
					if( term->ibuffer_disp == term->ibuffer_current ){
						eink_clear_lines( background_graphic, 0, term->buffer_current->row, DEFAULT_ATTR );
					}
					terminal_clear( term, term->ibuffer_current, 0, 0, term->buffer_current->column, term->buffer_current->row );
					term->cursor.x = term->cursor.y = 0;
					break;
				}
				break;
			case 'K': // Erase in Line
				if( arg[0] == -1 ) arg[0] = 0;
				switch( arg[0] ){
				case 0: // Clear from cursor to end of line
					if( term->ibuffer_disp == term->ibuffer_current ){
						eink_clear_chars( background_graphic, term->cursor.x, term->cursor.y, 
										term->buffer_current->column - term->cursor.x, DEFAULT_ATTR );
					}
					terminal_clear( term, term->ibuffer_current, term->cursor.x, term->cursor.y, 
											term->buffer[ term->ibuffer_current ].column, term->cursor.y );
					break;
				case 1: // Clear from cursor to beginning of line
					if( term->ibuffer_disp == term->ibuffer_current ){
						eink_clear_chars( background_graphic, 0, term->cursor.y, term->cursor.x, DEFAULT_ATTR );
					}
					terminal_clear( term, term->ibuffer_current, 0, term->cursor.y, 
											term->cursor.x, term->cursor.y );
					//term->cursor.x = 0;
					break;
				case 2: // Clear entire line
					if( term->ibuffer_disp == term->ibuffer_current ){
						eink_clear_chars( background_graphic, 0, term->cursor.y, term->buffer_current->column, DEFAULT_ATTR );
					}
					terminal_clear( term, term->ibuffer_current, 0, term->cursor.y, 
											term->buffer_current->column, term->cursor.y );
					term->cursor.x = 0;
					break;
				}
				break;
			case 's': // Save Cursor Position
				saved_cursor = term->cursor;
				break;
			case 'u': // Restore Cursor Position
				term->cursor = saved_cursor;
				break;
			case 'm': // Select Graphic Rendition
				if( arg[0] == -1 ) arg[0] = 0;
				{
					int ari;
					for( ari = 0; ari < 3; ari ++ ){
						switch( arg[ari] ){
						case -1:	break;
						case 0:		*attr = DEFAULT_ATTR;									break;
						case 7:		*attr |=  ATTR_INVERSE;									break;
						case 27:	*attr &= ~ATTR_INVERSE;									break;
						case 4:		*attr |=  ATTR_UNDERLINE;								break;
						case 24:	*attr &= ~ATTR_UNDERLINE;								break;
						case 1:		*attr |=  ATTR_BOLD;									break;
						case 22:	*attr &= ~ATTR_BOLD;									break;
						case 30:	*attr = ATTR( 15, ATTR_BG(*attr) );						break;
						case 31:	*attr = ATTR( 13, ATTR_BG(*attr) );						break;
						case 32:	*attr = ATTR( 12, ATTR_BG(*attr) );						break;
						case 33:	*attr = ATTR( 10, ATTR_BG(*attr) );						break;
						case 34:	*attr = ATTR( 9, ATTR_BG(*attr) );						break;
						case 35:	*attr = ATTR( 7, ATTR_BG(*attr) );						break;
						case 36:	*attr = ATTR( 6, ATTR_BG(*attr) );						break;
						case 37:	*attr = ATTR( 4, ATTR_BG(*attr) );						break;
						case 39:	*attr = ATTR( ATTR_FG(DEFAULT_ATTR), ATTR_BG(*attr) );	break;
						case 40:	*attr = ATTR( ATTR_FG(*attr), 15 );						break;
						case 41:	*attr = ATTR( ATTR_FG(*attr), 13 );						break;
						case 42:	*attr = ATTR( ATTR_FG(*attr), 12 );						break;
						case 43:	*attr = ATTR( ATTR_FG(*attr), 10 );						break;
						case 44:	*attr = ATTR( ATTR_FG(*attr), 9 );						break;
						case 45:	*attr = ATTR( ATTR_FG(*attr), 7 );						break;
						case 46:	*attr = ATTR( ATTR_FG(*attr), 6 );						break;
						case 47:	*attr = ATTR( ATTR_FG(*attr), 0 );						break;
						case 49:	*attr = ATTR( ATTR_FG(*attr), ATTR_BG(DEFAULT_ATTR) );	break;
						}
					}
				}
				break;
			case 'c': // Device Code Report
				if( qc == '>' ){
					sprintf( ret, "\033[>0;95;0c" );
				}else{
					sprintf( ret, "\033[?1;2c" );
				}
				write( pty, &ret, strlen( ret ) );
				break;
			case 'n': // VT100 Device Status
				fsync( pty );
				switch( arg[0] ){
				case 5: // Device Status Report
					sprintf( ret, "\033[0n" ); // OK
					//sprintf( ret, "\033[3n" ); // Failure
					write( pty, &ret, strlen( ret ) );
					break;
				case 6: // Cursor Position Report
					sprintf( ret, "\033[%d;%dR", term->cursor.y + 1, term->cursor.x + 1 );
					// printf("\n--- return\n" );
					write( pty, &ret, strlen( ret ) );
					break;
				}
				fsync( pty );
				break;
			case 'r': // Set Scrolling Region
				if( arg[0] <= 0 || arg[1] <= 0 ){
					arg[0] = 1;
					arg[1] = term->buffer_current->row;
				}
				term->buffer_current->scroll_top = arg[ 0 ] - 1;
				term->buffer_current->scroll_bottom = arg[ 1 ] - 1;
				break;
			default:
//				printf(" %c  (%x),%d,%d,%d  %d,%d\n", cmd, cmd, arg[0], arg[1], arg[2], term->buffer_current->column, term->buffer_current->row ); fflush( stdout );
				break;
			}
			continue;
		}
		//	printf("%c", key ); fflush( stdout );

		if( 0x20 <= key ){
			if( term->cursor.x >= term->buffer_current->column ){
				term->cursor.x = 0;
				//term->cursor.y ++;
				if( term->buffer_current->scroll_bottom + 1 != term->buffer_current->row || term->buffer_current->scroll_top != 0 ){
					if( term->cursor.y >= term->buffer_current->scroll_bottom ){
						terminal_delete_line( term,  term->ibuffer_current, term->buffer_current->scroll_top, 
													term->cursor.y + 1 - term->buffer_current->scroll_bottom );
						terminal_refreash_screen( background_graphic, term );
						
						term->cursor.y = term->buffer_current->scroll_bottom;
						//break;
					}else{
						term->cursor.y += 1;
					}
				}else{
					term->cursor.y += 1;
				}
			}

			if( term->cursor.y >= term->buffer_current->row ){
				if( !term->decpm[ DECPM_UseAlternateScreenBuffer ] ){
					if( ( term->ibuffer_current + 1 ) % term->nbuffer == term->ibuffer_disp ){
						term->ibuffer_current = term->ibuffer_disp;
					}
					if( term->ibuffer_current == term->ibuffer_disp ){
						eink_flush( 1, 0 );

						term->ibuffer_disp ++;
						term->ibuffer_disp %= term->nbuffer;
						term->buffer_disp = &term->buffer[ term->ibuffer_disp ];
					}
					term->ibuffer_current ++;
					term->ibuffer_current %= term->nbuffer;
					term->buffer_current = &term->buffer[ term->ibuffer_current ];
					terminal_clear( term, term->ibuffer_current, 0, 0, 
							term->buffer[ term->ibuffer_current ].column - 1, term->buffer[ term->ibuffer_current ].row - 1 );
					eink_clear_lines( background_graphic, 0, term->buffer_current->row, *attr );
					term->cursor.y = 0;
				}else{
					term->cursor.y = term->buffer_current->row - 1;
				}
			}
			/*if( key == '\t' ){
				eink_clear_chars( background_graphic, term->cursor.x, term->cursor.y, 1, *attr );
				term->cursor.x = ( term->cursor.x / 8 + 1 ) * 8;
				if( term->cursor.x >= term->buffer_current->column ){
					if( term->decpm[ DECPM_WraparoundMode ] && !term->decpm[ DECPM_UseAlternateScreenBuffer ] ){
						term->cursor.y ++;
						term->cursor.x = 0;
					}else{
						term->cursor.x = term->buffer_current->column - 1;
					}
				}
			}*/
			if( validate_pos( term, &term->cursor ) ){
				if( term->setmode[ SM_InsertMode ] ){
					if( key >= 0x80 ){
						terminal_insert_char( term,  term->ibuffer_current, term->cursor.y, term->cursor.x, 2 );
					}else{
						terminal_insert_char( term,  term->ibuffer_current, term->cursor.y, term->cursor.x, 1 );
					}
				}
				if( term->ibuffer_current == term->ibuffer_disp ){
					eink_putc( background_graphic, term->cursor.x, term->cursor.y, key, *attr );
				}
				terminal_setchar( term, term->ibuffer_current, term->cursor.x, term->cursor.y, 
												key, *attr );
				if( term->setmode[ SM_InsertMode ] ){
					terminal_refreash_line( background_graphic, term, term->cursor.y );
				}
			}
			term->cursor.x ++;
			if( key >= 0x80 ){
				if( term->cursor.x < term->buffer_current->column ){
					terminal_setchar( term, term->ibuffer_current, term->cursor.x, term->cursor.y, 
													0, *attr );
				}
				term->cursor.x ++;
			}
		}
	}
	term->cursor.x = term->cursor.x;
	term->cursor.y = term->cursor.y;
	//terminal_refreash_screen( background_graphic, term );
	
	return flush;
}

void terminal_refreash_line( char *background_graphic, struct terminal *term, int row )
{
	int i;
	
	eink_clear_lines( background_graphic, row, 1, DEFAULT_ATTR );
	for( i = 0; i < term->buffer_disp->column; i ++ ){
		if( term->buffer_disp->text[ row* term->buffer_current->column + i ] ){
			eink_putc( background_graphic, i, row, 
				term->buffer_disp->text[ row * term->buffer_current->column + i ], 
				term->buffer_disp->attr[ row * term->buffer_current->column + i ] );
		}
	}
//	if( term->decpm[DECPM_ShowCursor] && term->ibuffer_current == term->ibuffer_disp &&
//			validate_pos( term, &term->cursor ) )
//		eink_cursor( background_graphic, term->cursor.x, term->cursor.y, 1, DEFAULT_ATTR );
}

void terminal_refreash_screen( char *background_graphic, struct terminal *term )
{
	int i, j;
	
	eink_clear_lines( background_graphic, 0, term->buffer_disp->row, DEFAULT_ATTR );
	for( j = 0; j < term->buffer_disp->row; j ++ ){
		for( i = 0; i < term->buffer_disp->column; i ++ ){
			if( term->buffer_disp->text[ j * term->buffer_current->column + i ] ){
				eink_putc( background_graphic, i, j, 
					term->buffer_disp->text[ j * term->buffer_current->column + i ], 
					term->buffer_disp->attr[ j * term->buffer_current->column + i ] );
			}
		}
	}
	if( term->decpm[DECPM_ShowCursor] && term->ibuffer_current == term->ibuffer_disp &&
			validate_pos( term, &term->cursor ) )
		eink_cursor( background_graphic, term->cursor.x, term->cursor.y, 1, DEFAULT_ATTR );
	terminal_print_title( background_graphic, term );
}

void terminal_print_cursor( char *background_graphic, struct terminal *term, attr_t attr )
{
	static struct position cursor_hist = { 0, 0 };
	static attr_t attr_hist;
	
	if( validate_pos( term, &cursor_hist ) ){
		eink_cursor( background_graphic, cursor_hist.x, cursor_hist.y, 0, attr_hist );
	}
	if( term->decpm[ DECPM_ShowCursor ] && term->ibuffer_current == term->ibuffer_disp && 
			validate_pos( term, &term->cursor ) ){
		eink_cursor( background_graphic, term->cursor.x, term->cursor.y, term->decpm[ DECPM_ShowCursor ], attr );
	}
	cursor_hist = term->cursor;
	attr_hist = attr_hist;
}

void terminal_print_selected( char *background_graphic, struct terminal *term, struct position startpos, struct position endpos, char bar )
{
	static int startpos_hist = 0;
	static int endpos_hist = 0;
	int i;
	int s, e;
	int min, max;
	
	s = startpos.y * term->buffer_current->column + startpos.x;
	e =   endpos.y * term->buffer_current->column +   endpos.x;
	if( s > e ){
		int tmp;
		tmp = s;
		s = e;
		e = tmp;
	}
	min = minimum( s, startpos_hist );
	max = maximum( e, endpos_hist );
	
	for( i = 0; i <= term->buffer_current->row * term->buffer_current->column; i ++ ){
		int x, y;
		attr_t attr;
		
		x = i % term->buffer_current->column;
		y = i / term->buffer_current->column;
		
		if( s <= i && i <= e ){
		//	eink_cursor( background_graphic, x, y, bar, terminal_getattr( term, term->ibuffer_current, x, y ) );
			attr = term->buffer_disp->attr[y * term->buffer_current->column + x ];
			if( bar ) attr = ATTR( ATTR_BG(attr), ATTR_FG(attr) );
			eink_putc( background_graphic, x, y, 
				term->buffer_disp->text[ y * term->buffer_current->column + x ], attr );
		}else if( startpos_hist <= i && i <= endpos_hist &&
					( s != startpos_hist || e != endpos_hist ) ){
			attr = term->buffer_disp->attr[y * term->buffer_current->column + x ];
			eink_putc( background_graphic, x, y, 
				term->buffer_disp->text[ y * term->buffer_current->column + x ], attr );
		//	eink_cursor( background_graphic, x, y, 0, terminal_getattr( term, term->ibuffer_current, x, y ) );
		}
	}
	startpos_hist = s;
	endpos_hist = e;
}
