#ifndef __TERMINAL_H__
#define __TERMINAL_H__

//#define TERMINAL_DEBUG
	
#define INBUFFER_SIZE	(8192*2)

typedef unsigned short		attr_t;

#define ATTR(a,b)		(((a&0xF)<<4)|(b&0xF))
#define ATTR_FG(a)		((a&0xF0)>>4)
#define ATTR_BG(a)		((a&0xF))

#define DEFAULT_ATTR	ATTR(15,0)


#define ATTR_INVERSE	0x0100
#define ATTR_UNDERLINE	0x0200
#define ATTR_BOLD		0x0400


enum decpm{
	DECPM_ApplicationCursorKeys = 1,
	DECPM_DesignateUSASCIIforcharactersetsG0toG3 = 2,
	DECPM_132ColumnMode = 3,
	DECPM_SmoothScroll = 4,
	DECPM_ReverseVideo = 5,
	DECPM_OriginMode = 6,
	DECPM_WraparoundMode = 7,
	DECPM_AutorepeatKeys = 8,
	DECPM_SendMouseXandYonButtonPress = 9,
	DECPM_ShowToolbar = 10,
	DECPM_StartBlinkingCursor = 12,
	DECPM_PrintFormFeed = 18,
	DECPM_SetPrintExtentToFullscreen = 19,
	DECPM_ShowCursor = 25,
	DECPM_ShowScrollbar = 30,
	DECPM_EnableFontShiftingFunctions = 35,
	DECPM_EnterTektronixMode = 38,
	DECPM_Allow80to132Mode = 40,
	DECPM_Morefix = 41,
	DECPM_EnableNationReplacementCharactersets = 42,
	DECPM_TurnOnMarginBell = 44,
	DECPM_ReverseWraparoundMode = 45,
	DECPM_StartLogging = 46,
	DECPM_UseAlternateScreenBuffer = 47,
	DECPM_ApplicationKeypad = 66,
	DECPM_BackarrowkeySendsBackspace = 67,
	DECPM_MAX,
	DECPM_UseNormalScreenBufferAndRestoreCursor = 1049,
	DECPM_SaveCursorAndUseAlternateScreenBuffer = 1049
};

enum SM{
	SM_KeyboardActionMode = 2,
	SM_InsertMode = 4,
	SM_SendReceive = 12,
	SM_AutomaticNewline = 20,
	SM_MAX
};

struct position{
	int x;
	int y;
};

struct terminal_buffer{
	int column;
	int row;
	int scroll_top;
	int scroll_bottom;
	unsigned long *text;
	attr_t *attr;
};

#define TITLE_LENGTH	128

struct terminal{
	struct terminal_buffer *buffer_disp;
	struct terminal_buffer *buffer_current;
	struct terminal_buffer *buffer;
	int nbuffer;
	int ibuffer_disp;
	int ibuffer_current;
	struct position cursor;
	unsigned char decpm[ DECPM_MAX ];
	unsigned char setmode[ SM_MAX ];
	char title[ TITLE_LENGTH ];
};

void terminal_init( struct terminal *term, int nbuffer, int width, int height );
void terminal_reset( struct terminal *term );
void terminal_nextpage( struct terminal *term );
void terminal_prevpage( struct terminal *term );
void terminal_print_selected( char *background_graphic, struct terminal *term, struct position startpos, struct position endpos, char bar );
void terminal_print_cursor( char *background_graphic, struct terminal *term, attr_t attr );
void terminal_refreash_screen( char *background_graphic, struct terminal *term );
int terminal_input( char *background_graphic, struct terminal *term, int pty, attr_t *attr );
int validate_pos( struct terminal *term, struct position *pos );
void attr_extract( attr_t attr, unsigned char *fg, unsigned char *bg, unsigned char *under );
void terminal_refreash_line( char *background_graphic, struct terminal *term, int row );
void terminal_reinit( struct terminal *term, int nbuffer, int width, int height );
void terminal_free( struct terminal *term );


#endif
